@@ -20,10 +20,10 @@
 //! # Missing features that could be added
 //!
 //! * Options to _not_ copy mtimes or permissions.
-//! * Continue copying after an error.
-//! * Callbacks for progress or logging, error handling, etc.
+//! * A callback that can decide whether to continue after an error.
 //! * Overwrite existing directories or files.
 //! * Copy single files: don't assume the source path is a directory.
+//! * A dry-run mode.
 //!
 //! # Example
 //!
@@ -42,6 +42,27 @@
 //!
 //! # Release history
 //!
+//! ## 0.4.0
+//!
+//! Released 2021-11-30
+//!
+//! API changes:
+//!
+//! * Remove `copy_buffer_size`, `file_buffers_copied`: these are too niche to have in the public
+//!   API, and anyhow become meaningless when we use [std::fs::copy].
+//!
+//! * New [ErrorKind::DestinationDoesNotExist].
+//!
+//! * [Error::io_error] returns `Option<io::Error>`: errors from this crate may not have a
+//!   direct `io::Error` source.
+//!
+//! * [CopyOptions::copy_tree] arguments are relaxed to `AsRef<Path>` so that they will accept
+//!   `&str`, `PathBuf`, `tempfile::TempDir`, etc.
+//!
+//! Improvements:
+//!
+//! * Use [std::fs::copy], which is more efficient, and makes this crate simpler.
+//!
 //! ## 0.3.1
 //!
 //! Released 2021-11-07
@@ -92,20 +113,18 @@
 use std::collections::VecDeque;
 use std::fmt;
 use std::fs::{self, DirEntry};
-use std::io::{self, Read, Write};
+use std::io;
 use std::path::{Path, PathBuf};
 
 /// Options for copying file trees.
 ///
 /// Default options may be OK for many callers:
 /// * Preserve mtime and permissions.
-/// * Use an 8MiB copy buffer.
 /// * Create the destination if it does not exist.
 pub struct CopyOptions<'f> {
     // TODO: Continue or stop on error?
     // TODO: Option controlling whether to copy mtimes?
     // TODO: Copy permissions?
-    copy_buffer_size: usize,
     create_destination: bool,
 
     // I agree with Clippy that the callbacks are complex types, but stable Rust
@@ -121,7 +140,6 @@ pub struct CopyOptions<'f> {
 impl<'f> Default for CopyOptions<'f> {
     fn default() -> CopyOptions<'f> {
         CopyOptions {
-            copy_buffer_size: 8 << 20,
             create_destination: true,
             filter: None,
             after_entry_copied: None,
@@ -135,15 +153,6 @@ impl<'f> CopyOptions<'f> {
         CopyOptions::default()
     }
 
-    /// Set the buffer size for copying regular files.
-    pub fn copy_buffer_size(self, copy_buffer_size: usize) -> CopyOptions<'f> {
-        assert!(copy_buffer_size > 0);
-        CopyOptions {
-            copy_buffer_size,
-            ..self
-        }
-    }
-
     /// Set whether to create the destination if it does not exist (the default), or return an error.
     ///
     /// Only the immediate destination is created, not all its parents.
@@ -217,29 +226,35 @@ impl<'f> CopyOptions<'f> {
     /// Copy the tree according to the options.
     ///
     /// Returns [CopyStats] describing how many files were copied, etc.
-    pub fn copy_tree(mut self, src: &Path, dest: &Path) -> Result<CopyStats> {
+    pub fn copy_tree<P, Q>(mut self, src: P, dest: Q) -> Result<CopyStats>
+    where
+        P: AsRef<Path>,
+        Q: AsRef<Path>,
+    {
+        let src = src.as_ref();
+        let dest = dest.as_ref();
+
         let mut stats = CopyStats::default();
 
-        if self.create_destination && !dest.is_dir() {
-            copy_dir(src, dest, &mut stats)?;
+        // TODO: Handle the src not being a dir: copy that single entry.
+        if self.create_destination {
+            if !dest.is_dir() {
+                copy_dir(src, dest, &mut stats)?;
+            }
+        } else if !dest.is_dir() {
+            return Err(Error::new(ErrorKind::DestinationDoesNotExist, dest));
         }
 
         let mut subdir_queue: VecDeque<PathBuf> = VecDeque::new();
         subdir_queue.push_back(PathBuf::from(""));
-        assert!(self.copy_buffer_size > 0);
-        let mut copy_buf = vec![0u8; self.copy_buffer_size];
 
         while let Some(subdir) = subdir_queue.pop_front() {
             let subdir_full_path = src.join(&subdir);
-            for entry in fs::read_dir(&subdir_full_path).map_err(|io| Error {
-                path: subdir_full_path.clone(),
-                io,
-                kind: ErrorKind::ReadDir,
-            })? {
-                let dir_entry = entry.map_err(|io| Error {
-                    path: subdir_full_path.clone(),
-                    io,
-                    kind: ErrorKind::ReadDir,
+            for entry in fs::read_dir(&subdir_full_path)
+                .map_err(|io| Error::from_io_error(io, ErrorKind::ReadDir, &subdir_full_path))?
+            {
+                let dir_entry = entry.map_err(|io| {
+                    Error::from_io_error(io, ErrorKind::ReadDir, &subdir_full_path)
                 })?;
                 let entry_subpath = subdir.join(dir_entry.file_name());
                 if let Some(filter) = &mut self.filter {
@@ -250,24 +265,19 @@ impl<'f> CopyOptions<'f> {
                 }
                 let src_fullpath = src.join(&entry_subpath);
                 let dest_fullpath = dest.join(&entry_subpath);
-                let file_type = dir_entry.file_type().map_err(|io| Error {
-                    path: src_fullpath.clone(),
-                    io,
-                    kind: ErrorKind::ReadDir,
-                })?;
+                let file_type = dir_entry
+                    .file_type()
+                    .map_err(|io| Error::from_io_error(io, ErrorKind::ReadDir, &src_fullpath))?;
                 if file_type.is_file() {
-                    copy_file(&src_fullpath, &dest_fullpath, &mut copy_buf, &mut stats)?
+                    copy_file(&src_fullpath, &dest_fullpath, &mut stats)?
                 } else if file_type.is_dir() {
                     copy_dir(&src_fullpath, &dest_fullpath, &mut stats)?;
                     subdir_queue.push_back(entry_subpath.clone());
                 } else if file_type.is_symlink() {
                     copy_symlink(&src_fullpath, &dest_fullpath, &mut stats)?
                 } else {
-                    return Err(Error::new(
-                        ErrorKind::UnsupportedFileType,
-                        src_fullpath,
-                        format!("unsupported file type {:?}", file_type),
-                    ));
+                    // TODO: Include the file type.
+                    return Err(Error::new(ErrorKind::UnsupportedFileType, src_fullpath));
                 }
                 if let Some(ref mut f) = self.after_entry_copied {
                     f(&entry_subpath, &file_type, &stats);
@@ -289,10 +299,6 @@ pub struct CopyStats {
     pub symlinks: usize,
     /// The number of bytes of file content copied, across all files.
     pub file_bytes: u64,
-    /// The number of file buffers copied, per [CopyOptions::copy_buffer_size].
-    ///
-    /// (This is fairly obscure and mostly intended for testing.)
-    pub file_buffer_reads: usize,
     /// The number of entries filtered out by the [CopyOptions::filter] callback.
     pub filtered_out: usize,
 }
@@ -300,28 +306,40 @@ pub struct CopyStats {
 /// An error from copying a tree.
 ///
 /// At present this library does not support continuing after an error, so only the first error is
-/// returned.
+/// returned by [CopyOptions::copy_tree].
 #[derive(Debug)]
 pub struct Error {
     path: PathBuf,
-    io: io::Error,
+    /// The original IO error, if any.
+    io: Option<io::Error>,
     kind: ErrorKind,
 }
 
-/// A [std::result::Result] possibly containing a cp_r [Error].
+/// A [std::result::Result] possibly containing a `cp_r` [Error].
 pub type Result<T> = std::result::Result<T, Error>;
 
 impl Error {
-    /// Construct a new error.
-    pub fn new(kind: ErrorKind, path: PathBuf, message: String) -> Error {
-        let io_kind: io::ErrorKind = match kind {
-            ErrorKind::UnsupportedFileType => io::ErrorKind::Unsupported,
-            other => unimplemented!("unhandled {:?}", other),
-        };
+    /// Construct a new error with no source.
+    pub fn new<P>(kind: ErrorKind, path: P) -> Error
+    where
+        P: Into<PathBuf>,
+    {
         Error {
-            path,
+            path: path.into(),
             kind,
-            io: io::Error::new(io_kind, message),
+            io: None,
+        }
+    }
+
+    /// Construct a new error from a [std::io::Error].
+    pub fn from_io_error<P>(io: io::Error, kind: ErrorKind, path: P) -> Error
+    where
+        P: Into<PathBuf>,
+    {
+        Error {
+            path: path.into(),
+            kind,
+            io: Some(io),
         }
     }
 
@@ -331,10 +349,9 @@ impl Error {
         &self.path
     }
 
-    /// The IO error that caused this error, or a description of this error as an IO error, if it
-    /// was not directly caused by one.
-    pub fn io_error(&self) -> &io::Error {
-        &self.io
+    /// The IO error that caused this error, if any.
+    pub fn io_error(&self) -> Option<&io::Error> {
+        self.io.as_ref()
     }
 
     /// The kind of error that occurred.
@@ -345,7 +362,13 @@ impl Error {
 
 impl std::error::Error for Error {
     fn source(&self) -> Option<&(dyn std::error::Error + 'static)> {
-        Some(&self.io)
+        // It seems like you should be able to spell this like `self.io.as_ref().into()` but that
+        // doesn't work and I'm not sure why...
+        if let Some(io) = &self.io {
+            Some(io)
+        } else {
+            None
+        }
     }
 }
 
@@ -360,8 +383,14 @@ impl fmt::Display for Error {
             ReadSymlink => "reading symlink",
             CreateSymlink => "creating symlink",
             UnsupportedFileType => "unsupported file type",
+            CopyFile => "copying file",
+            DestinationDoesNotExist => "destination directory does not exist",
         };
-        write!(f, "{}: {}: {}", kind_msg, self.path.display(), self.io)
+        if let Some(io) = &self.io {
+            write!(f, "{}: {}: {}", kind_msg, self.path.display(), io)
+        } else {
+            write!(f, "{}: {}", kind_msg, self.path.display())
+        }
     }
 }
 
@@ -375,6 +404,8 @@ pub enum ErrorKind {
     ReadFile,
     /// Error creating or writing a destination file.
     WriteFile,
+    /// Error in copying a file: might be a read or write error.
+    CopyFile,
     /// Error creating a destination directory.
     CreateDir,
     /// Error reading a symlink.
@@ -384,90 +415,40 @@ pub enum ErrorKind {
     /// The source tree contains a type of file that this library can't copy, such as a Unix
     /// FIFO.
     UnsupportedFileType,
+    /// The destination directory does not exist.
+    DestinationDoesNotExist,
 }
 
-fn copy_file(src: &Path, dest: &Path, buf: &mut [u8], stats: &mut CopyStats) -> Result<()> {
-    let mut inf = fs::OpenOptions::new()
-        .read(true)
-        .open(src)
-        .map_err(|io| Error {
-            kind: ErrorKind::ReadFile,
-            path: src.to_owned(),
-            io,
-        })?;
-    let mut outf = fs::OpenOptions::new()
-        .create_new(true)
-        .write(true)
-        .open(dest)
-        .map_err(|io| Error {
-            kind: ErrorKind::WriteFile,
-            path: dest.to_owned(),
-            io,
-        })?;
-    loop {
-        let len = inf.read(buf).map_err(|io| Error {
-            kind: ErrorKind::ReadFile,
-            path: src.to_owned(),
-            io,
-        })?;
-        if len == 0 {
-            break;
-        }
-        stats.file_bytes += len as u64;
-        stats.file_buffer_reads += 1;
-        outf.write_all(&buf[..len]).map_err(|io| Error {
-            kind: ErrorKind::WriteFile,
-            path: dest.to_owned(),
-            io,
-        })?;
-    }
+fn copy_file(src: &Path, dest: &Path, stats: &mut CopyStats) -> Result<()> {
+    // TODO: Optionally first check and error if the destination exists.
+    let bytes_copied =
+        fs::copy(src, dest).map_err(|io| Error::from_io_error(io, ErrorKind::CopyFile, src))?;
+    stats.file_bytes += bytes_copied;
 
-    let inf_metadata = inf.metadata().map_err(|io| Error {
-        kind: ErrorKind::ReadFile,
-        path: src.to_owned(),
-        io,
-    })?;
-
-    let src_mtime = filetime::FileTime::from_last_modification_time(&inf_metadata);
-    filetime::set_file_handle_times(&outf, None, Some(src_mtime)).map_err(|io| Error {
-        kind: ErrorKind::WriteFile,
-        path: dest.to_owned(),
-        io,
-    })?;
-
-    outf.set_permissions(inf_metadata.permissions())
-        .map_err(|io| Error {
-            kind: ErrorKind::WriteFile,
-            path: dest.to_owned(),
-            io,
-        })?;
+    let src_metadata = src
+        .metadata()
+        .map_err(|io| Error::from_io_error(io, ErrorKind::ReadFile, src))?;
+    let src_mtime = filetime::FileTime::from_last_modification_time(&src_metadata);
+    filetime::set_file_mtime(&dest, src_mtime)
+        .map_err(|io| Error::from_io_error(io, ErrorKind::WriteFile, dest))?;
 
+    // Permissions should have already been set by fs::copy.
     stats.files += 1;
     Ok(())
 }
 
 fn copy_dir(_src: &Path, dest: &Path, stats: &mut CopyStats) -> Result<()> {
     fs::create_dir(dest)
-        .map_err(|io| Error {
-            kind: ErrorKind::CreateDir,
-            path: dest.to_owned(),
-            io,
-        })
+        .map_err(|io| Error::from_io_error(io, ErrorKind::CreateDir, dest))
         .map(|()| stats.dirs += 1)
 }
 
 #[cfg(unix)]
 fn copy_symlink(src: &Path, dest: &Path, stats: &mut CopyStats) -> Result<()> {
-    let target = fs::read_link(src).map_err(|io| Error {
-        kind: ErrorKind::ReadSymlink,
-        path: src.to_owned(),
-        io,
-    })?;
-    std::os::unix::fs::symlink(target, dest).map_err(|io| Error {
-        kind: ErrorKind::CreateSymlink,
-        path: dest.to_owned(),
-        io,
-    })?;
+    let target =
+        fs::read_link(src).map_err(|io| Error::from_io_error(io, ErrorKind::ReadSymlink, src))?;
+    std::os::unix::fs::symlink(target, dest)
+        .map_err(|io| Error::from_io_error(io, ErrorKind::CreateSymlink, dest))?;
     stats.symlinks += 1;
     Ok(())
 }
